@@ -2,14 +2,16 @@ package AnyEvent::Redis;
 
 use strict;
 use 5.008_001;
-our $VERSION = '0.22';
+our $VERSION = '0.23';
 
 use constant DEBUG => $ENV{ANYEVENT_REDIS_DEBUG};
 use AnyEvent;
 use AnyEvent::Handle;
 use AnyEvent::Socket;
+use AnyEvent::Redis::Protocol;
 use Try::Tiny;
 use Carp qw(croak);
+use Encode ();
 
 our $AUTOLOAD;
 
@@ -19,6 +21,11 @@ sub new {
     my $host = delete $args{host} || '127.0.0.1';
     my $port = delete $args{port} || 6379;
 
+    if (my $encoding = $args{encoding}) {
+        $args{encoding} = Encode::find_encoding($encoding);
+        croak qq{Encoding "$encoding" not found} unless ref $args{encoding};
+    }
+
     bless {
         host => $host,
         port => $port,
@@ -63,7 +70,8 @@ sub connect {
 
     my $cv;
     if (@_) {
-        $cv = AE::cv;
+        $cv = pop if UNIVERSAL::isa($_[-1], 'AnyEvent::CondVar');
+        $cv ||= AE::cv;
         push @{$self->{connect_queue}}, [ $cv, @_ ];
     }
 
@@ -78,6 +86,8 @@ sub connect {
               return
             };
 
+        binmode $fh; # ensure bytes until we decode
+
         my $hd = AnyEvent::Handle->new(
             fh => $fh,
             on_error => sub { $_[0]->destroy;
@@ -88,6 +98,7 @@ sub connect {
             on_eof   => sub { $_[0]->destroy;
                               $self->cleanup('connection closed');
                           },
+            encoding => $self->{encoding},
         );
 
         $self->{cmd_cb} = sub {
@@ -99,16 +110,18 @@ sub connect {
                 $cv = pop if UNIVERSAL::isa($_[-1], 'AnyEvent::CondVar');
                 $cb = pop if ref $_[-1] eq 'CODE';
             }
+            $cv ||= AE::cv;
 
             my $send = join("\r\n",
                   "*" . (1 + @_),
-                  map(('$' . length $_ => $_), uc($command), @_))
+                  map { ('$' . length $_ => $_) }
+                        (uc($command), map { $self->{encoding} && length($_)
+                                             ? $self->{encoding}->encode($_)
+                                             : $_ } @_))
                 . "\r\n";
 
             warn $send if DEBUG;
 
-            $cv ||= AE::cv;
-
             $hd->push_write($send);
 
             # Are we already subscribed to anything?
@@ -132,7 +145,7 @@ sub connect {
                     }
                 }) if $cb;
 
-                $hd->push_read(ref $self => sub {
+                $hd->push_read("AnyEvent::Redis::Protocol" => sub {
                         my($res, $err) = @_;
 
                         if($command eq 'info') {
@@ -144,7 +157,7 @@ sub connect {
 
                         $self->all_cv->end;
                         $err ? $cv->croak($res) : $cv->send($res);
-                    });
+                });
 
             } else {
                 croak "Must provide a CODE reference for subscriptions" unless $cb;
@@ -154,7 +167,7 @@ sub connect {
 
                 my $res_cb; $res_cb = sub {
 
-                    $hd->push_read(ref $self => sub {
+                    $hd->push_read("AnyEvent::Redis::Protocol" => sub {
                             my($res, $err) = @_;
 
                             if(ref $res) {
@@ -204,125 +217,6 @@ sub connect {
     return $cv;
 }
 
-sub anyevent_read_type {
-    my(undef, $cb) = @_;
-
-    sub {
-        my($hd) = @_;
-
-        return unless defined $hd->{rbuf};
-
-        if($hd->{rbuf} =~ /^[-+:]/) {
-            $hd->{rbuf} =~ s/^([-+:])([^\015\012]*)\015?\012// or return;
-
-            $cb->($2, $1 eq '-');
-
-            return 1;
-
-        } elsif($hd->{rbuf} =~ /^\$/) {
-            $hd->{rbuf} =~ s/^\$([-0-9]+)\015?\012// or return;
-            my $len = $1;
-
-            if($len < 0) {
-                $cb->(undef);
-            } elsif($len + 2 <= length $hd->{rbuf}) {
-                $cb->(substr $hd->{rbuf}, 0, $len);
-                # Remove ending newline
-                substr $hd->{rbuf}, 0, $len + 2, "";
-            } else {
-                $hd->unshift_read (chunk => $len + 2, sub {
-                        $cb->(substr $_[1], 0, $len);
-                    });
-            }
-
-            return 1;
-
-        } elsif($hd->{rbuf} =~ /^\*/) {
-
-            $hd->{rbuf} =~ s/^\*([-0-9]+)\015?\012// or return;
-            my $size = $1;
-            warn "expecting $size values" if DEBUG;
-
-            my @lines;
-
-            my $reader; $reader = sub {
-                my($hd, $async) = @_;
-
-                while(@lines < $size) {
-                    if($hd->{rbuf} =~ /^([\$\-+:])([^\012\015]+)\015?\012/) {
-                        my $type = $1;
-                        my $line = $2;
-
-                        if($type =~ /[-+:]/) {
-                            $hd->{rbuf} =~ s/^[^\012\015]+\015?\012//;
-                            push @lines, $line;
-                        } elsif($line < 0) {
-                            $hd->{rbuf} =~ s/^[^\012\015]+\015?\012//;
-                            push @lines, undef;
-
-                        } elsif((1 + length($2) + 2 # Initial line
-                                 + $line + 2)       # Data and newline
-                                <= length $hd->{rbuf}) {
-                            $hd->{rbuf} =~ s/^[^\012\015]+\015?\012//;
-                            push @lines, substr $hd->{rbuf}, 0, $line, "";
-                            $hd->{rbuf} =~ s/^\015?\012//;
-
-                        } else {
-                            # Data not buffered, so we need to do this async
-                            $hd->unshift_read($reader) if $async;
-                            return $async;
-                        }
-                    } elsif($hd->{rbuf} =~ /^\*/) { # Nested
-                        my $reader_closure = $reader; # Need to avoid holding circular ref
-
-                        $hd->unshift_read(__PACKAGE__, sub {
-                                push @lines, $_[0];
-
-                                if(@lines == $size) {
-                                    warn "$size nested values" if DEBUG;
-                                    $cb->(\@lines);
-                                    undef $reader_closure;
-                                } else {
-                                    $hd->unshift_read($reader_closure);
-                                }
-                                return 1;
-                            });
-
-                        undef $reader;
-                        return 1;
-                    } else {
-                        $hd->unshift_read($reader) if $async;
-                        return $async;
-                    }
-                }
-
-                if($size < 0 || @lines == $size) {
-                    warn "Got $size values" if DEBUG;
-                    $cb->($size < 0 ? undef : \@lines);
-                    undef $reader;
-                    return 1;
-                }
-
-                die "unreachable";
-            };
-
-            return $reader->($hd, 1);
-
-        } elsif(length $hd->{rbuf}) {
-            # remove extra lines
-            $hd->{rbuf} =~ s/^\015?\012//g;
-
-            if(length $hd->{rbuf}) {
-                # Unknown
-                $cb->("Unknown type", 1);
-                return 1;
-            }
-        }
-
-        return;
-    }
-}
-
 1;
 __END__
 
@@ -341,6 +235,7 @@ AnyEvent::Redis - Non-blocking Redis client
   my $redis = AnyEvent::Redis->new(
       host => '127.0.0.1',
       port => 6379,
+      encoding => 'utf8',
       on_error => sub { warn @_ },
   );
 
@@ -358,7 +253,38 @@ AnyEvent::Redis - Non-blocking Redis client
 
 =head1 DESCRIPTION
 
-AnyEvent::Redis is a non-blocking Redis client using AnyEvent.
+AnyEvent::Redis is a non-blocking (event-driven) Redis client.
+
+This module is an AnyEvent user; you must install and use a supported event loop.
+
+=head1 ESTABLISHING A CONNECTION
+
+To create a new connection, use the new() method with the following attributes:
+
+=over
+
+=item host => <HOSTNAME>
+
+B<Required.>  The hostname or literal address of the server.  
+
+=item port => <PORT>
+
+Optional.  The server port.
+
+=item encoding => <ENCODING>
+
+Optional.  Encode and decode data (when storing and retrieving, respectively)
+according to I<ENCODING> (C<"utf8"> is recommended or see L<Encode::Supported>
+for details on possible I<ENCODING> values).
+
+Omit if you intend to handle raw binary data with this connection.
+
+=item on_error => $cb->($errmsg)
+
+Optional.  Callback that will be fired if a connection or database-level error
+occurs.  The error message will be passed to the callback as the sole argument.
+
+=back
 
 =head1 METHODS
 
@@ -366,7 +292,7 @@ All methods supported by your version of Redis should be supported.
 
 =head2 Normal commands
 
-There are two alternative approaches for handling results from commands.
+There are two alternative approaches for handling results from commands:
 
 =over 4
 
@@ -376,11 +302,21 @@ There are two alternative approaches for handling results from commands.
     # arguments to command
   );
 
+  # Then...
+  my $res;
+  eval { 
+      # Could die()
+      $res = $cv->recv;
+  }; 
+  warn $@ if $@;
+
+  # or...
   $cv->cb(sub {
     my($cv) = @_;
     my($result, $err) = $cv->recv
   });
 
+
 =item * Callback:
 
   $redis->command(
@@ -425,8 +361,7 @@ connection with an active subscription are subscribe and unsubscribe commands.
 
 =back
 
-The Redis command reference
-(L<http://code.google.com/p/redis/wiki/CommandReference>) lists all commands
+The Redis command reference (L<http://redis.io/commands>) lists all commands
 Redis supports.
 
 =head1 REQUIREMENTS
@@ -460,6 +395,8 @@ Jeremy Zawodny
 
 Leon Brocard
 
+Michael S. Fischer
+
 =head1 SEE ALSO
 
 L<Redis>, L<AnyEvent>
